from spetlr import Configurator

from . import extras
from .DeliverySqlServerSpn import DeliverySqlServerSpn
from .test_deliverysql import DeliverySqlServerTests


class DeliverySqlServerSpnTests(DeliverySqlServerTests):
    @classmethod
    def setUpClass(cls):
        cls.sql_server = DeliverySqlServerSpn()
        cls.tc = Configurator()

        cls.tc.add_resource_path(extras)
        cls.tc.reset(debug=True)



from spetlrtools.testing import DataframeTestCase

from spetlr.exceptions import OnlyUseInSpetlrDebugMode
from spetlr.testutils.stop_test_streams import stop_test_streams


class TestStopTestStreams(DataframeTestCase):
    def test_01_configurator_must_be_debug(self):
        with self.assertRaises(OnlyUseInSpetlrDebugMode):
            stop_test_streams()



import pyspark.sql.types as T
from spetlrtools.testing import DataframeTestCase

from spetlr.exceptions import AmbiguousColumnsAfterCleaning
from spetlr.spark import Spark
from spetlr.transformers import CleanColumnNamesTransformer


class TestCleanColumnNamesTransformer(DataframeTestCase):
    """
    Testing the CleanColumnNamesTransformer.
    """

    input_schema = T.StructType(
        [
            T.StructField("col (col)", T.StringType(), True),
            T.StructField("col ", T.StringType(), True),
            T.StructField("col three", T.StringType(), True),
            T.StructField("col-four", T.StringType(), True),
            T.StructField("col5", T.StringType(), True),
            T.StructField("col<6", T.StringType(), True),
            T.StructField("col.seven", T.StringType(), True),
            T.StructField("..eight.?!", T.StringType(), True),
            T.StructField("??", T.StringType(), True),
            T.StructField("colx@# $%^&*()", T.StringType(), True),
            T.StructField(" col space ", T.StringType(), True),
            T.StructField("col_123_456", T.StringType(), True),
            T.StructField("x", T.StringType(), True),
        ]
    )
    input_data = []

    def test_01_clean_column_names_transformer(self):
        # Creating an input dataframe using the specified schema

        input_df = Spark.get().createDataFrame(
            data=self.input_data, schema=self.input_schema
        )

        # Transforming the input data
        df_transformed = CleanColumnNamesTransformer().process(df=input_df)

        # Specifying the expected schema
        expected_schema = T.StructType(
            [
                T.StructField("col_col", T.StringType(), True),
                T.StructField("col", T.StringType(), True),
                T.StructField("col_three", T.StringType(), True),
                T.StructField("col_four", T.StringType(), True),
                T.StructField("col5", T.StringType(), True),
                T.StructField("col6", T.StringType(), True),
                T.StructField("colseven", T.StringType(), True),
                T.StructField("eight", T.StringType(), True),
                T.StructField("", T.StringType(), True),
                T.StructField("colx", T.StringType(), True),
                T.StructField("col_space", T.StringType(), True),
                T.StructField("col_123_456", T.StringType(), True),
                T.StructField("x", T.StringType(), True),
            ]
        )

        # Creating expected synthetic data
        expected_data = []

        # Creating a dataframe with the expected output using the specified schema
        df_expected = Spark.get().createDataFrame(expected_data, expected_schema)

        # Performing the Equality test
        self.assertEqualSchema(df_transformed.schema, df_expected.schema)
        self.assertDataframeMatches(df=df_transformed, expected_data=expected_data)

    def test_02_clean_column_names_transformer_exclude_one_column(self):
        # Creating an input dataframe using the specified schema
        input_df = Spark.get().createDataFrame(
            data=self.input_data, schema=self.input_schema
        )

        # Transforming the input data
        df_transformed = CleanColumnNamesTransformer(
            exclude_columns=["col (col)"]
        ).process(df=input_df)

        # Specifying the expected schema
        expected_schema = T.StructType(
            [
                T.StructField("col (col)", T.StringType(), True),
                T.StructField("col", T.StringType(), True),
                T.StructField("col_three", T.StringType(), True),
                T.StructField("col_four", T.StringType(), True),
                T.StructField("col5", T.StringType(), True),
                T.StructField("col6", T.StringType(), True),
                T.StructField("colseven", T.StringType(), True),
                T.StructField("eight", T.StringType(), True),
                T.StructField("", T.StringType(), True),
                T.StructField("colx", T.StringType(), True),
                T.StructField("col_space", T.StringType(), True),
                T.StructField("col_123_456", T.StringType(), True),
                T.StructField("x", T.StringType(), True),
            ]
        )

        # Creating expected synthetic data
        expected_data = []

        # Creating a dataframe with the expected output using the specified schema
        df_expected = Spark.get().createDataFrame(expected_data, expected_schema)

        # Performing the Equality test
        self.assertEqualSchema(df_transformed.schema, df_expected.schema)
        self.assertDataframeMatches(df=df_transformed, expected_data=expected_data)

    def test_03_clean_column_names_transformer_exclude_multiple_columns(self):
        # Creating an input dataframe using the specified schema
        input_df = Spark.get().createDataFrame(
            data=self.input_data, schema=self.input_schema
        )

        # Transforming the input data
        df_transformed = CleanColumnNamesTransformer(
            exclude_columns=["col (col)", "col-four", "??"]
        ).process(df=input_df)

        # Specifying the expected schema
        expected_schema = T.StructType(
            [
                T.StructField("col (col)", T.StringType(), True),
                T.StructField("col", T.StringType(), True),
                T.StructField("col_three", T.StringType(), True),
                T.StructField("col-four", T.StringType(), True),
                T.StructField("col5", T.StringType(), True),
                T.StructField("col6", T.StringType(), True),
                T.StructField("colseven", T.StringType(), True),
                T.StructField("eight", T.StringType(), True),
                T.StructField("??", T.StringType(), True),
                T.StructField("colx", T.StringType(), True),
                T.StructField("col_space", T.StringType(), True),
                T.StructField("col_123_456", T.StringType(), True),
                T.StructField("x", T.StringType(), True),
            ]
        )

        # Creating expected synthetic data
        expected_data = []

        # Creating a dataframe with the expected output using the specified schema
        df_expected = Spark.get().createDataFrame(expected_data, expected_schema)

        # Performing the Equality test
        self.assertEqualSchema(df_transformed.schema, df_expected.schema)
        self.assertDataframeMatches(df=df_transformed, expected_data=expected_data)

    def test_04_ambiguous_columns(self):
        input_schema = T.StructType(
            [
                T.StructField("col ", T.StringType(), True),
                T.StructField("col  ", T.StringType(), True),
            ]
        )
        input_data = []

        # Creating an input dataframe using the specified schema
        input_df = Spark.get().createDataFrame(data=input_data, schema=input_schema)

        # Transforming the input data that should raise error
        with self.assertRaises(AmbiguousColumnsAfterCleaning):
            CleanColumnNamesTransformer().process(df=input_df)



import uuid
import warnings
from abc import abstractmethod
from datetime import datetime
from functools import reduce
from typing import List

import pyspark.sql.functions as F
from pyspark.sql import DataFrame
from pyspark.sql.types import StringType, TimestampType

from spetlr.etl import Transformer
from spetlr.etl.types import dataset_group


class LogTransformer(Transformer):
    """
    Base class for specific ETL transformers that log measurements on a DataFrame.

    The `LogTransformer` class serves as a foundation for implementing various
    transformers that take measurements on a DataFrame. These measurements can
    include quantitative and qualitative values, such as counting rows,
    counting nulls, or any other custom metric that the user wishes to log
    during the ETL process. A log transformer produces a dataframe with a single line
    of data (measurement). It is recommended to use the log transformer with the
    LogOrchestrator class, so the load process is automatically dealt with.

    Usage:
        The `LogTransformer` class should be subclassed to create specific
        transformers that measure and log various metrics during the ETL
        process. Implement the `log` and `log_many` methods to define the
        specific measurement logic.

    Arguments:
        log_name (str): The name for the log entry.
        dataset_input_keys (List[str], optional): A list of input dataset keys
            that this transformer will process. Defaults to None.
        dataset_output_key (str, optional): The key for the output dataset
            generated by this transformer. If not provided, a random UUID will
            be used as the default value. It is not required for when using the
            LogOrchestrator.
        consume_inputs (bool, optional): Flag to control dataset consuming behavior.
            Defaults to True

    Methods:
        log(df: DataFrame) -> DataFrame:
            Abstract method that should be implemented in the derived classes.
            This method takes a DataFrame and logs the desired measurement(s).

        log_many(datasets: dataset_group) -> DataFrame:
            Abstract method that should be implemented in the derived classes.
            This method takes a dataset of dataframes and logs the desired
            measurement(s). Defaults to logging each frame by its `log` method.
    """

    def __init__(
        self,
        *,
        log_name: str,
        dataset_input_keys: List[str] = None,
        consume_inputs: bool = True,
    ):
        super().__init__(
            dataset_input_keys=dataset_input_keys,
            dataset_output_key=uuid.uuid4(),  # to ensure unique log output key
            consume_inputs=consume_inputs,
        )
        self.log_name = log_name
        self.timestamp_now = datetime.utcnow()
        self.log_method_name = self.__class__.__name__

    @abstractmethod
    def log(self, df: DataFrame) -> DataFrame:
        raise NotImplementedError()

    # This is the default method for log many, if the logging in the subclass can
    # be applied to each dataframe in the dataset separately. However, if the logging
    # logic should be performed across multiple dataframes in the dataset, then the
    # user should implement their own method.
    @abstractmethod
    def log_many(self, dataset: dataset_group) -> DataFrame:
        dfs_to_log = [
            self.log(df).withColumn("DatasetInputKey", F.lit(key))
            for key, df in dataset.items()
        ]

        df_log = reduce(
            lambda df1, df2: df1.unionByName(df2),
            dfs_to_log,
        )

        return df_log

    def _create_log(self, df_log: DataFrame) -> DataFrame:
        # some columns are casted to ensure the same types upon return

        df_log_cols = df_log.columns

        for col in [
            "LogName",
            "LogTimestamp",
            "LogMethodName",
            "LogId",
        ]:
            if col in df_log_cols:
                warnings.warn(
                    f"The Log Transformer used ({self.log_method_name}) produced a"
                    + f"column named '{col}'. Consider that this will be overwritten by"
                    + "the _create_log() method."
                )

        df = (
            df_log.withColumn("LogName", F.lit(self.log_name).cast(StringType()))
            .withColumn("LogTimestamp", F.lit(self.timestamp_now).cast(TimestampType()))
            .withColumn("LogMethodName", F.lit(self.log_method_name).cast(StringType()))
        )

        # this creates an id per log transformer
        df = df.withColumn("LogId", F.lit(str(uuid.uuid4())))

        columns_final = [
            "LogId",
            "LogName",
            "LogTimestamp",
            "LogMethodName",
        ] + df_log.columns

        df = df.select(*columns_final)

        return df

    def process(self, df: DataFrame) -> DataFrame:
        df = self.log(df)
        df = df.withColumn("DatasetInputKey", F.lit(self.dataset_input_keys[0]))
        df = self._create_log(df)
        return df

    def process_many(self, datasets: dataset_group) -> DataFrame:
        df = self.log_many(datasets)
        df = self._create_log(df)
        return df



import json

from pyspark.sql.types import BinaryType, StructField, StructType, TimestampType
from spetlrtools.testing import DataframeTestCase
from spetlrtools.time import dt_utc

from spetlr import Configurator
from spetlr.delta import DeltaHandle
from spetlr.orchestrators.ehjson2delta.EhJsonToDeltaTransformer import (
    EhJsonToDeltaTransformer,
)
from spetlr.spark import Spark


class JsonEhTransformerUnitTests(DataframeTestCase):
    tc: Configurator
    capture_eventhub_output_schema = StructType(
        [
            StructField("Body", BinaryType(), True),
            StructField("pdate", TimestampType(), True),
            StructField("EnqueuedTimestamp", TimestampType(), True),
        ]
    )

    df_in = Spark.get().createDataFrame(
        [
            (
                json.dumps(
                    {
                        "id": 1234,
                        "name": "John",
                    }
                ).encode(
                    "utf-8"
                ),  # Body
                dt_utc(2021, 10, 31, 0, 0, 0),  # pdate
                dt_utc(2021, 10, 31, 0, 0, 0),  # EnqueuedTimestamp
            ),
        ],
        capture_eventhub_output_schema,
    )

    @classmethod
    def setUpClass(cls) -> None:
        cls.tc = Configurator()
        cls.tc.clear_all_configurations()
        cls.tc.set_debug()

        cls.tc.register("TblPdate1", {"name": "TablePdate1{ID}"})
        cls.tc.register("TblPdate2", {"name": "TablePdate2{ID}"})
        cls.tc.register("TblPdate3", {"name": "TablePdate3{ID}"})

        spark = Spark.get()

        spark.sql(f"DROP TABLE IF EXISTS {cls.tc.table_name('TblPdate1')}")
        spark.sql(f"DROP TABLE IF EXISTS {cls.tc.table_name('TblPdate2')}")
        spark.sql(f"DROP TABLE IF EXISTS {cls.tc.table_name('TblPdate3')}")

        spark.sql(
            f"""
                CREATE TABLE {cls.tc.table_name('TblPdate1')}
                (id int, name string, BodyJson string, pdate timestamp,
                EnqueuedTimestamp timestamp)
                PARTITIONED BY (pdate)
            """
        )

        spark.sql(
            f"""
                CREATE TABLE {cls.tc.table_name('TblPdate2')}
                (id int, name string, pdate timestamp, EnqueuedTimestamp timestamp)
                PARTITIONED BY (pdate)
            """
        )

        spark.sql(
            f"""
                CREATE TABLE {cls.tc.table_name('TblPdate3')}
                (id int, name string, pdate timestamp, EnqueuedTimestamp timestamp,
                Unknown string)
                PARTITIONED BY (pdate)
            """
        )

    def test_01_transformer_w_body(self):
        """Tests whether the body is saved as BodyJson"""

        dh = DeltaHandle.from_tc("TblPdate1")

        expected = [
            (
                1234,
                "John",
                dt_utc(2021, 10, 31, 0, 0, 0),
                dt_utc(2021, 10, 31, 0, 0, 0),
                json.dumps(
                    {
                        "id": 1234,
                        "name": "John",
                    }
                ),  # BodyJson is at the end of select statement
            ),
        ]

        df_result = EhJsonToDeltaTransformer(target_dh=dh).process(self.df_in)

        # Check that data is correct
        self.assertDataframeMatches(df_result, None, expected)

    def test_02_transformer(self):
        """Test if the data is correctly extracted"""
        dh = DeltaHandle.from_tc("TblPdate2")

        expected = [
            (
                1234,
                "John",
                dt_utc(2021, 10, 31, 0, 0, 0),
                dt_utc(2021, 10, 31, 0, 0, 0),
            ),
        ]

        df_result = EhJsonToDeltaTransformer(target_dh=dh).process(self.df_in)

        # Check that data is correct
        self.assertDataframeMatches(df_result, None, expected)

    def test_03_transformer_unknown_target_field(self):
        """This should test what happens if the target
        schema has a field that does not exist in the source dataframe."""
        dh = DeltaHandle.from_tc("TblPdate3")

        expected = [
            (
                1234,
                "John",
                # THe unknown column is extract as NONE here.
                # Since pdate and EnqueuedTimestamp is always added at the end.
                None,
                dt_utc(2021, 10, 31, 0, 0, 0),
                dt_utc(2021, 10, 31, 0, 0, 0),
            ),
        ]

        df_result = EhJsonToDeltaTransformer(target_dh=dh).process(self.df_in)

        # Check that data is correct
        self.assertDataframeMatches(df_result, None, expected)



"""
Obtain a singleton spark instance.
Add configuration options using the method Spark.configure.
Obtain the instance by calling Spark.get()
Some standard options are pre-set, call configure with value=None to remove them.
"""

# This class uses module level singleton pattern as suggested by method5 of
# https://stackoverflow.com/questions/6760685/creating-a-singleton-in-python


from typing import Optional, Tuple

from pyspark.sql import SparkSession


class Spark:
    """
    A class singleton to get a valid Spark session
    """

    _spark: Optional[SparkSession] = None
    _configurations = {
        "spark.sql.autoBroadcastJoinThreshold": -1,
        "spark.sql.session.timeZone": "Etc/UTC",
        "spark.driver.extraJavaOptions": "-Duser.timezone=UTC",
        "spark.executor.extraJavaOptions": "-Duser.timezone=UTC",
    }
    _master = None

    @classmethod
    def master(cls, master: str):
        cls._master = master
        return cls

    @classmethod
    def config(cls, key, value=None) -> None:
        """
        :param key: The spark configuration key
        :param value: The value to set.
            If the value is missing or None, the configurarion will be removed
        :return:
        """
        if cls._spark is not None:
            raise Exception("Configuration method called after spark session build.")
        if value is not None:
            cls._configurations[key] = value
        else:
            if key in cls._configurations:
                del cls._configurations[key]
        return cls

    @classmethod
    def get(cls) -> SparkSession:
        """
        :return:
        The current spark session.
        """
        if cls._spark is not None:
            return cls._spark
        builder = SparkSession.builder
        if cls._master is not None:
            builder = builder.master(cls._master)
        for key, value in cls._configurations.items():
            builder = builder.config(key, value)
        cls._spark = builder.getOrCreate()
        return cls._spark

    @classmethod
    def version(cls) -> Tuple:
        return tuple(int(p) for p in cls.get().version.split("."))

    DATABRICKS_RUNTIME_9_1 = (3, 1, 2)
    DATABRICKS_RUNTIME_11_3 = (3, 3, 0)
    DATABRICKS_RUNTIME_10_4 = (3, 2, 1)



from typing import List

from pyspark.sql import DataFrame

from spetlr.etl import Transformer
from spetlr.exceptions import AmbiguousColumnsAfterCleaning


class CleanColumnNamesTransformer(Transformer):
    """
    Clean up columns names by removing special characters,
    trimming whitespaces and changing other whitespaces to underscores

    NB: This transformer throws exception if ambiguous column names are created

    Parameters
    ----------
        exclude_columns : Potential columns to exclude
    """

    def __init__(
        self,
        exclude_columns: List[str] = None,
        dataset_input_keys: List[str] = None,
        dataset_output_key: str = None,
        consume_inputs: bool = True,
    ):
        super().__init__(
            dataset_input_keys=dataset_input_keys,
            dataset_output_key=dataset_output_key,
            consume_inputs=consume_inputs,
        )
        self.exclude_columns = exclude_columns or []
        self.keep_chars = [" ", "_", "-"]  # Special characters not removed
        self.replace_chars = [" ", "-"]  # Characters that are replaced with "_"

    def process(self, df: DataFrame) -> DataFrame:
        columns = [col for col in df.columns if col not in self.exclude_columns]
        for col in columns:
            new_col_name = "".join(
                e for e in col if e.isalnum() or e in self.keep_chars
            )
            new_col_name = new_col_name.strip()
            for c in self.replace_chars:
                new_col_name = new_col_name.replace(c, "_")
            df = df.withColumnRenamed(col, new_col_name)

        if len(set(df.columns)) != len(df.columns):
            raise AmbiguousColumnsAfterCleaning(
                "Column names after cleaning are ambiguous!"
            )

        return df



"""
Specifications are based on this documentation:
https://learn.microsoft.com/en-us/azure/databricks/sql/language-manual/sql-ref-syntax-ddl-create-table-using
"""

from spetlr.deltaspec.DeltaTableSpec import DeltaTableSpec
from spetlr.deltaspec.DeltaTableSpecBase import DeltaTableSpecBase
from spetlr.deltaspec.DeltaTableSpecDifference import DeltaTableSpecDifference
from spetlr.deltaspec.exceptions import (
    InvalidSpecificationError,
    NoTableAtTarget,
    TableSpecNotReadable,
    TableSpecSchemaMismatch,
    TableSpectNotEnforcable,
)



import argparse
import json
from types import SimpleNamespace

from spetlr.exceptions import SpetlrException
from spetlr.functions import init_dbutils


def main():
    parser = argparse.ArgumentParser(description="spetlr mountpoint setup.")
    parser.add_argument(
        "path",
        nargs="+",
        type=str,
        help="The json file to read the configuration from.\n"
        """Example json contents:
[
  {
    "storageAccountName":"spetlr",
    "secretScope":"spetlr",
    "clientIdName":"Databricks--ClientId",
    "clientSecretName":"Databricks--ClientSecret",
    "tenantIdName":"Databricks--TenantId",
    "containers":["silver"]
  }
]
        """,
    )
    args = parser.parse_args()

    required_keys = [
        "storageAccountName",
        "secretScope",
        "clientIdName",
        "clientSecretName",
        "tenantIdName",
        "containers",
    ]
    configs = []
    for path in args.path:
        with open(path) as json_file:
            for item in json.load(json_file):
                for key in required_keys:
                    if key not in item:
                        raise SpetlrException(f"{path} contains invalid shape.")
                configs.append(SimpleNamespace(**item))

    dbutils = init_dbutils()

    # Mounting
    all_mounts = [m[0] for m in dbutils.fs.mounts()]

    for item in configs:
        scope = item.secretScope

        clientId = dbutils.secrets.get(scope=scope, key=item.clientIdName)
        # ServicePrincipal KEY
        clientSecret = dbutils.secrets.get(scope=scope, key=item.clientSecretName)
        tenantId = dbutils.secrets.get(scope=scope, key=item.tenantIdName)

        prfx = "fs.azure.account."
        configs = {
            f"{prfx}auth.type": "OAuth",
            f"{prfx}oauth.provider.type": (
                "org.apache.hadoop.fs.azurebfs.oauth2.ClientCredsTokenProvider"
            ),
            f"{prfx}oauth2.client.id": clientId,
            f"{prfx}oauth2.client.secret": clientSecret,
            f"{prfx}oauth2.client.endpoint": (
                f"https://login.microsoftonline.com/{tenantId}/oauth2/token"
            ),
        }

        storageAccountUrl = item.storageAccountName + ".dfs.core.windows.net/"

        # Mount each container from the storage accounts
        for container in item.containers:
            source = f"abfss://{container}@" + storageAccountUrl
            mount_name = f"/mnt/{item.storageAccountName}/{container}/"

            if mount_name not in all_mounts:
                print(f"Mounting {mount_name}")
                dbutils.fs.mount(
                    source=source, mount_point=mount_name, extra_configs=configs
                )
            else:
                print(f"{mount_name} is already mounted")

    dbutils.fs.refreshMounts()


if __name__ == "__main__":
    main()



from typing import List

from pyspark.sql import DataFrame

from spetlr.etl import Loader, dataset_group
from spetlr.etl.loaders import SimpleLoader
from spetlr.etl.loaders.stream_loader import StreamLoader
from spetlr.tables import TableHandle


class UpsertLoaderStreaming(Loader):
    def __init__(
        self,
        handle: TableHandle,
        options_dict: dict = None,
        trigger_type: str = None,
        trigger_time_seconds: int = None,
        query_name: str = None,
        checkpoint_path: str = None,
        await_termination: bool = False,
        upsert_join_cols: List[str] = None,
    ):
        super().__init__()

        self._loader = StreamLoader(
            loader=SimpleLoader(
                handle=handle, join_cols=upsert_join_cols, mode="upsert"
            ),
            options_dict=options_dict,
            trigger_type=trigger_type,
            trigger_time_seconds=trigger_time_seconds,
            outputmode="update",
            query_name=query_name,
            checkpoint_path=checkpoint_path,
            await_termination=await_termination,
        )

    def save_many(self, datasets: dataset_group) -> None:
        raise NotImplementedError()

    def save(self, df: DataFrame) -> None:
        """Upserts a single dataframe to the target table."""
        self._loader.save(df)



import unittest

from pyspark.sql import functions as f
from pyspark.sql.dataframe import DataFrame
from spetlrtools.time import dt_utc

from spetlr import Configurator
from spetlr.cache import CachedLoader, CachedLoaderParameters
from spetlr.delta import DbHandle, DeltaHandle
from spetlr.spark import Spark


class ChildCacher(CachedLoader):
    to_be_written: DataFrame
    written: DataFrame
    to_be_deleted: DataFrame
    deleted: DataFrame

    def write_operation(self, df: DataFrame):
        self.to_be_written = df
        self.written = df.filter(df["b"].isin([1, 2]))
        DeltaHandle.from_tc("CachedTestTarget").append(self.written)
        return self.written.withColumn("myId", f.lit(12345))

    def delete_operation(self, df: DataFrame) -> DataFrame:
        target_name = Configurator().table_name("CachedTestTarget")

        self.to_be_deleted = df
        Spark.get().sql(f"DELETE FROM {target_name} WHERE b = 8")
        self.deleted = df.filter(df["b"] == 8)
        return self.deleted


class CachedLoaderTests(unittest.TestCase):
    params: CachedLoaderParameters
    old_cache = [
        (
            "3",
            3,
            453652661,
            dt_utc(2022, 1, 1, 14),
            None,
            99,
        ),  # match
        (
            "6",
            6,
            123456789,
            dt_utc(2022, 1, 1, 14),
            None,
            99,
        ),  # mismatch
        (
            "7",
            7,
            1284583559,
            dt_utc(2022, 1, 1, 14),
            None,
            99,
        ),  # match
        (
            "8",
            8,
            123456789,
            dt_utc(2022, 1, 1, 14),
            None,
            99,
        ),  # unknown row, to be deleted
        (
            "9",
            9,
            123456789,
            dt_utc(2022, 1, 1, 14),
            None,
            99,
        ),  # unknown row, to be deleted
    ]

    new_data = [
        ("1", 1, "foo1"),  # new
        ("2", 2, "foo2"),  # new
        ("3", 3, "foo3"),  # match
        ("6", 6, "foo6"),  # mismatch
        ("7", 7, "foo7"),  # match
        ("7", 7, "foo7"),  # duplicate row will only be loaded once
    ]

    @classmethod
    def setUpClass(cls) -> None:
        tc = Configurator()
        tc.clear_all_configurations()
        tc.set_debug()

        tc.register("TestDb", dict(name="test{ID}", path="/tmp/test{ID}.db"))
        tc.register(
            "CachedTest",
            dict(
                name="test{ID}.cachedloader_cache",
                path="/tmp/test{ID}.db/cachedloader_cache",
            ),
        )
        tc.register(
            "CachedTestTarget",
            dict(
                name="test{ID}.cachedloader_target",
                path="/tmp/test{ID}.db/cachedloader_target",
            ),
        )
        DbHandle.from_tc("TestDb").create()
        spark = Spark.get()
        spark.sql(
            """
            CREATE TABLE IF NOT EXISTS {CachedTest_name}
            (
                a STRING,
                b INTEGER,
                rowHash INTEGER,
                loadedTime TIMESTAMP,
                deletedTime TIMESTAMP,
                myId INTEGER
            )
            USING DELTA
            COMMENT "Caching Test"
            LOCATION "{CachedTest_path}"
        """.format(
                **tc.get_all_details()
            )
        )

        spark.sql(
            """
            CREATE TABLE IF NOT EXISTS {CachedTestTarget_name}
            (
                a STRING,
                b INTEGER,
                payload STRING
            )
            USING DELTA
            COMMENT "Caching target"
            LOCATION "{CachedTestTarget_path}"
        """.format(
                **tc.get_all_details()
            )
        )

        cls.params = CachedLoaderParameters(
            cache_table_name=tc.table_name("CachedTest"),
            key_cols=["a", "b"],
            cache_id_cols=["myId"],
        )

        cls.sut = ChildCacher(cls.params)

    @classmethod
    def tearDownClass(cls) -> None:
        DbHandle("TestDb").drop_cascade()

    def test_01_can_perform_cached_write(self):
        cache_dh = DeltaHandle.from_tc("CachedTest")
        # prime the cache
        df_old_cache = Spark.get().createDataFrame(
            self.old_cache, schema=cache_dh.read().schema
        )
        cache_dh.overwrite(df_old_cache)

        # prepare the new data
        target_dh = DeltaHandle.from_tc("CachedTestTarget")
        df_new = Spark.get().createDataFrame(
            self.new_data, schema=target_dh.read().schema
        )

        # execute the system under test
        self.sut.save(df_new)

        cache = cache_dh.read().withColumn(
            "isRecent",
            (f.current_timestamp().cast("long") - f.col("loadedTime").cast("long"))
            < 100,
        )
        cache.show()

        # Section on writing

        # We expect that the rows that were requested for write were
        # - new rows 1 & 2
        # - mismatch row 6
        to_be_written_ids = {row.a for row in self.sut.to_be_written.collect()}
        self.assertEqual(to_be_written_ids, {"1", "2", "6"})
        # filtered actually written
        written_ids = {row.a for row in self.sut.written.collect()}
        self.assertEqual(written_ids, {"1", "2"})
        # verify that this agrees with the current state of the cache:
        new_cache = cache.filter(f.col("isRecent"))
        new_cache_ids = {row.a for row in new_cache.collect()}
        self.assertEqual(new_cache_ids, {"1", "2"})

        # Section on deleting
        # to be deleted were the missing rows 8 and 9
        to_be_deleted_ids = {row.a for row in self.sut.to_be_deleted.collect()}
        self.assertEqual(to_be_deleted_ids, {"8", "9"})

        # only row 8 was passed to be deleted. Check.
        del_cache = cache.filter(cache[self.sut.params.deletedTime].isNotNull())
        (del_id,) = [row.a for row in del_cache.collect()]
        self.assertEqual(del_id, "8")



from typing import List

from pyspark.sql import DataFrame
from pyspark.sql.functions import concat_ws, md5


def Md5HashColumn(
    df: DataFrame, colName: str, cols_to_exclude: List[str] = None
) -> DataFrame:
    """
    Generate column with md5 encoding, based on a list of column.

    it's optional to add a list of columns to ignore when generating the column

    concat_ws replace null values with an empty string

    param df: The pyspark DataFrame where the column is added
    param colName: Name for the column
    param cols_to_exclude: Optional list of columns to exclude

    return: DataFrame with md5 hash column
    """

    cols_to_exclude = cols_to_exclude or []

    cols = [col for col in df.columns if col not in cols_to_exclude]
    df = df.withColumn(colName, md5(concat_ws("||", *cols)))
    return df



import warnings
from typing import List

from pyspark.sql import DataFrame

from spetlr.etl.transformer import Transformer
from spetlr.etl.types import dataset_group
from spetlr.utils.Md5HashColumn import Md5HashColumn


class DataChangeCaptureTransformer(Transformer):
    """
    This class archives data changes from a source table, using a primary key column.
    It identifies and captures new or modified records.
    For accurate comparison of all columns, ensure both schemas are identical.

    """

    def __init__(
        self,
        primary_key,
        df_source="source",
        df_target="target",
        cols_to_exclude: List[str] = None,
    ):
        self.primary_key = primary_key
        self.df_source = df_source
        self.df_target = df_target
        self.cols_to_exclude = cols_to_exclude or []
        if primary_key in self.cols_to_exclude:
            raise ValueError(f"The primary key '{primary_key}' should not be excluded.")

    def _warn_if_schema_differs(self, df1: DataFrame, df2: DataFrame):
        if df1.schema != df2.schema:
            warnings.warn(
                "Schemas are not identical, "
                "making the MD5 hash comparison potentially unreliable"
            )

    def process_many(self, dataset: dataset_group) -> DataFrame:
        df_source = dataset[self.df_source]
        df_target = dataset[self.df_target]

        self._warn_if_schema_differs(df_source, df_target)

        # Add an md5 hash column to both DataFrames
        df_source = Md5HashColumn(df_source, "md5_source", self.cols_to_exclude)
        df_target = Md5HashColumn(df_target, "md5_target", self.cols_to_exclude)
        df_joined = df_source.join(
            df_target.select("md5_target", self.primary_key),
            self.primary_key,
            "left_outer",
        )

        # Filter to keep rows where the MD5 hash values are different
        # or the target values are null
        df = df_joined.filter(
            (df_joined.md5_source != df_joined.md5_target)
            | df_joined.md5_target.isNull()
        ).drop("md5_source", "md5_target")
        return df



from typing import List

from spetlrtools.testing import DataframeTestCase

from spetlr import Configurator
from spetlr.delta import DbHandle, DeltaHandle
from spetlr.etl.loaders import SimpleLoader
from spetlr.utils import DataframeCreator
from tests.cluster.delta import extras
from tests.cluster.delta.SparkExecutor import SparkSqlExecutor


class SimpleLoaderUpsertTests(DataframeTestCase):
    target_id = "UpsertLoaderDummy"

    join_cols = ["col1", "col2"]

    data1 = [
        (5, 6, "foo"),
        (7, 8, "bar"),
    ]
    data2 = [
        (1, 2, "baz"),
    ]
    data3 = [(5, 6, "boo"), (5, 7, "spam")]
    # data5 is the merge result of data2 + data3 + data4
    data4 = [(1, 2, "baz"), (5, 6, "boo"), (5, 7, "spam"), (7, 8, "bar")]

    dummy_columns: List[str] = ["col1", "col2", "col3"]

    dummy_schema = None
    target_dh_dummy: DeltaHandle = None

    @classmethod
    def setUpClass(cls) -> None:
        Configurator().add_resource_path(extras)
        Configurator().set_debug()

        cls.target_dh_dummy = DeltaHandle.from_tc("UpsertLoaderDummy")

        SparkSqlExecutor().execute_sql_file("upsertloader-test")

        cls.dummy_schema = cls.target_dh_dummy.read().schema

        # make sure target is empty
        df_empty = DataframeCreator.make_partial(cls.dummy_schema, [], [])
        cls.target_dh_dummy.overwrite(df_empty)

    @classmethod
    def tearDownClass(cls) -> None:
        DbHandle.from_tc("UpsertLoaderDb").drop_cascade()

    def test_01_can_perform_incremental_on_empty(self):
        loader = SimpleLoader(
            handle=self.target_dh_dummy, mode="upsert", join_cols=self.join_cols
        )

        df_source = DataframeCreator.make_partial(
            self.dummy_schema, self.dummy_columns, self.data1
        )

        loader.save(df_source)
        self.assertDataframeMatches(self.target_dh_dummy.read(), None, self.data1)

    def test_02_can_perform_incremental_append(self):
        """The target table is already filled from before."""
        existing_rows = self.target_dh_dummy.read().collect()
        self.assertEqual(2, len(existing_rows))

        loader = SimpleLoader(
            handle=self.target_dh_dummy, mode="upsert", join_cols=self.join_cols
        )

        df_source = DataframeCreator.make_partial(
            self.dummy_schema, self.dummy_columns, self.data2
        )

        loader.save(df_source)

        self.assertDataframeMatches(
            self.target_dh_dummy.read(), None, self.data1 + self.data2
        )

    def test_03_can_perform_merge(self):
        """The target table is already filled from before."""
        existing_rows = self.target_dh_dummy.read().collect()
        self.assertEqual(3, len(existing_rows))

        loader = SimpleLoader(
            handle=self.target_dh_dummy, mode="upsert", join_cols=self.join_cols
        )

        df_source = DataframeCreator.make_partial(
            self.dummy_schema, self.dummy_columns, self.data3
        )

        loader.save(df_source)

        self.assertDataframeMatches(self.target_dh_dummy.read(), None, self.data4)



from pyspark.sql import DataFrame

from spetlr.etl import Extractor
from spetlr.tables import TableHandle


class StreamExtractor(Extractor):
    """This extractor will extract from any object that has a .read_stream() method."""

    def __init__(self, handle: TableHandle, dataset_key: str = None):
        super().__init__(dataset_key=dataset_key)
        self.handle = handle

    def read(self) -> DataFrame:
        return self.handle.read_stream()



import unittest
from datetime import date
from unittest import mock

from freezegun import freeze_time
from pyspark.sql.types import (
    DateType,
    IntegerType,
    StringType,
    StructField,
    StructType,
    TimestampType,
)
from spetlrtools.time import dt_utc

from spetlr.azure_log_analytics import AzureLogAnalyticsHandle
from spetlr.utils import DataframeCreator


class TestAzureLogAnalyticsHandle(unittest.TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        cls.workspace_id = "4de5-8ic5"
        cls.resource = "/api/logs"

        schema = StructType(
            [
                StructField("col_1", IntegerType(), True),
                StructField("col_2", StringType(), True),
                StructField("col_3", TimestampType(), True),
                StructField("col_4", DateType(), True),
            ]
        )

        data = [
            (1, "first", dt_utc(2011, 1, 1, 8, 25, 25), date(2018, 3, 29)),
            (2, "second", dt_utc(2022, 2, 2, 18, 10, 10), date(2020, 2, 2)),
        ]

        cls.df = DataframeCreator.make(schema, data)

    def setUp(self) -> None:
        self.az_lah = AzureLogAnalyticsHandle(
            log_analytics_workspace_id=self.workspace_id,
            shared_key="",
        )

    def test_create_uri_01(self) -> None:
        uri = self.az_lah._create_uri(self.resource)

        expected_uri = (
            f"https://{self.workspace_id}.ods.opinsights"
            + f".azure.com{self.resource}?api-version=2016-04-01"
        )

        self.assertEqual(uri, expected_uri)

    def test_create_body_02(self) -> None:
        body = self.az_lah._create_body(self.df)

        expected_body = (
            """[\n"""
            + """    {\n"""
            + """        "col_1": 1,\n"""
            + """        "col_2": "first",\n"""
            + """        "col_3": "2011-01-01T08:25:25",\n"""
            + """        "col_4": "2018-03-29"\n"""
            + """    },\n"""
            + """    {\n"""
            + """        "col_1": 2,\n"""
            + """        "col_2": "second",\n"""
            + """        "col_3": "2022-02-02T18:10:10",\n"""
            + """        "col_4": "2020-02-02"\n"""
            + """    }\n"""
            + """]"""
        )

        self.assertEqual(body, expected_body)

    @freeze_time("2020-06-15 12:10:10")
    def test_create_headers_03(self) -> None:
        shared_key = "1/a=2b3-c4d+5e"

        az_lah_no_log_type = AzureLogAnalyticsHandle(
            log_analytics_workspace_id=self.workspace_id,
            shared_key=shared_key,
        )

        method = "POST"
        content_type = "application/json"
        content_length = 80
        resource = self.resource

        headers = az_lah_no_log_type._create_headers(
            method, content_type, content_length, resource
        )

        expected_headers = {
            "content-type": "application/json",
            "Authorization": "SharedKey 4de5-8ic5:uAALir2LY"
            + "1gju1qtYw8ut1bL4V9p3FR+0CxJhHjRkbY=",
            "Log-Type": "Databricks",
            "x-ms-date": "Mon, 15 Jun 2020 12:10:10 GMT",
        }

        self.assertEqual(headers, expected_headers)

    def test_http_data_collector_api_post_200_ok_04(self) -> None:
        with mock.patch("requests.post") as mock_requests_post:
            mock_response = mock.Mock()
            mock_response.status_code = 200
            mock_requests_post.return_value = mock_response

            self.assertEqual(self.az_lah.api_post(self.df), 200)

    def test_http_data_collector_api_post_400_warning_05(self) -> None:
        with mock.patch("requests.post") as mock_requests_post:
            mock_response = mock.Mock()
            mock_response.status_code = 400
            mock_requests_post.return_value = mock_response

            # when the response code from the api != 200, then it should warn
            with self.assertWarns(UserWarning):
                self.assertEqual(self.az_lah.api_post(self.df), 400)

    def test_append_06(self) -> None:
        # tests that the method 'append' is an alias to the 'api_post' method
        self.assertEqual(
            AzureLogAnalyticsHandle.__dict__["append"],
            AzureLogAnalyticsHandle.__dict__["api_post"],
        )

    def test_read_07(self) -> None:
        # tests that the method 'read' is an alias to the 'api_get' method
        self.assertEqual(
            AzureLogAnalyticsHandle.__dict__["read"],
            AzureLogAnalyticsHandle.__dict__["api_get"],
        )

        with self.assertRaises(NotImplementedError):
            self.az_lah.read()

    def test_prepare_dataframe_08(self) -> None:
        pass  # the function is implicitly tested in test_create_body_02


if __name__ == "__main__":
    unittest.main()



from typing import Dict, List, Union
from unittest.mock import MagicMock

from pyspark.sql import DataFrame

from spetlr.etl import Loader
from spetlr.etl.types import dataset_group


class MockLoader(Loader, MagicMock):
    def __init__(
        self, dataset_input_keys: Union[str, List[str]] = None, *args, **kwargs
    ):
        MagicMock.__init__(self, dataset_input_keys=dataset_input_keys, *args, **kwargs)
        self.saved: Dict[str, DataFrame] = {}

    def save(self, df: DataFrame) -> None:
        self.saved = {"single": df}

    def save_many(self, datasets: dataset_group) -> None:
        self.saved = datasets

    def getDf(self) -> DataFrame:
        return self.saved["single"]



from spetlr.configurator import Configurator
from spetlr.exceptions import OnlyUseInSpetlrDebugMode
from spetlr.spark import Spark


def stop_test_streams():
    """
    Stops all streams in the current SPETLR config session.

    This mean streams, that has a query name with the Configurator UUID
    """

    c = Configurator()
    if not c.is_debug():
        raise OnlyUseInSpetlrDebugMode()

    id_extension = c.get("ID")

    for stream in Spark.get().streams.active:
        if id_extension in stream.name:
            print(f'Stopping the stream "{stream.name}"')
            stream.stop()
            stream.awaitTermination()



from datetime import datetime, timezone

from pyspark.sql.types import (
    IntegerType,
    StringType,
    StructField,
    StructType,
    TimestampType,
)
from spetlrtools.testing import DataframeTestCase
from spetlrtools.time import dt_utc

from spetlr import Configurator
from spetlr.delta import DbHandle, DeltaHandle
from spetlr.etl import Orchestrator
from spetlr.etl.extractors import StreamExtractor
from spetlr.etl.loaders import StreamLoader
from spetlr.etl.loaders.scd2_loader import SCD2UpsertLoader
from spetlr.spark import Spark
from spetlr.testutils import stop_test_streams


class TestSCD2Loader(DataframeTestCase):
    @classmethod
    def setUpClass(cls) -> None:
        """
        Set up configuration for all test cases.
        Clearing any existing configurations and setting up the debug mode
        for better visibility during testing.
        """
        Configurator().clear_all_configurations()
        Configurator().set_debug()

    @classmethod
    def tearDownClass(cls) -> None:
        """
        Tear down the database setup after all tests have been run.
        Ensuring a clean state for subsequent tests.
        """
        DbHandle.from_tc("MyDb").drop_cascade()
        stop_test_streams()

    def test_01_single_join_col(self):
        """
        Test the SCD2UpsertLoader functionality.
        This involves setting up a test database and table,
        loading initial data, and then validating the
        Slowly Changing Dimension Type 2 (SCD2) behavior.
        """

        self._create_table()

        dh = DeltaHandle.from_tc("MyTbl")

        """
        First test takes this input data:
            | Id |   Col1   |    Col2    |      TimeCol      |
            |---------|----------|------------|------------------|
            |    1    |  Fender  | Telecaster | 2021-07-01 10:00 |
            |    1    |  Fender  | Telecaster | 2021-07-01 10:00 |
            |    1    |  Fender  | Telecaster | 2021-07-01 11:00 |
            |    2    |  Gibson  |  Les Paul  | 2021-07-01 11:00 |
            |    3    |  Ibanez  |     RG     | 2021-08-01 11:00 |
            |    3    |  Ibanez  |     RG     | 2021-09-01 11:00 |
            |    3    |  Ibanez  |     RG     | 2021-10-01 11:00 |


        """

        data_start = [
            (1, "Fender", "Telecaster", dt_utc(2021, 7, 1, 10)),
            # Duplicate when comparing the columns
            (1, "Fender", "Telecaster", dt_utc(2021, 7, 1, 10)),
            (1, "Fender", "Telecaster", dt_utc(2021, 7, 1, 11)),
            (2, "Gibson", "Les Paul", dt_utc(2021, 7, 1, 11)),
            (3, "Ibanez", "RG", dt_utc(2021, 8, 1, 11)),
            (3, "Ibanez", "RG", dt_utc(2021, 9, 1, 11)),
            (3, "Ibanez", "RG", dt_utc(2021, 10, 1, 11)),
        ]
        schema_start = StructType(
            [
                StructField("Id", IntegerType(), True),
                StructField("Col1", StringType(), True),
                StructField("Col2", StringType(), True),
                StructField("TimeCol", TimestampType(), True),
            ]
        )
        # Create a DataFrame from the test data
        df_in = Spark.get().createDataFrame(data=data_start, schema=schema_start)

        # Define the columns for joining and time column for the SCD2 process
        _join_cols = ["Id"]
        _time_col = "TimeCol"

        SCD2UpsertLoader(sink_handle=dh, join_cols=_join_cols, time_col=_time_col).save(
            df_in
        )

        max_time = datetime(2262, 4, 11, tzinfo=timezone.utc)

        """
        And generates this:
                |PK|Col1| Col2  |      Time     | ValFrom  |  ValTo   |Curr|Hash|
                |--|----|-------|---------------|----------|---------|----|----|
                |1 |Fend|Telec. |2021-07-01 10:|2021-07-01|21-07-01 1| F  |b69x|
                |1 |Fend|Telec. |2021-07-01 11:|2021-07-01| max_time | T  |276x|
                |2 |Gibs|L.Paul|2021-07-01 11:|2021-07-01| max_time | T  |f79x|
                |3 |Iban| RG   |2021-08-01 11:|2021-08-01|21-09-01 1| F  |7d5x|
                |3 |Iban| RG   |2021-09-01 11:|2021-09-01|21-10-01 1| F  |274x|
                |3 |Iban| RG   |2021-10-01 11:|2021-10-01| max_time | T  |fb1x|
        """

        # Define the expected outcome after the initial data load
        # This includes the handling of duplicate records and
        # setting up the initial state of the SCD2 table

        expected_data = [
            (
                1,
                "Fender",
                "Telecaster",
                dt_utc(2021, 7, 1, 10),
                dt_utc(2021, 7, 1, 10),
                dt_utc(2021, 7, 1, 11),
                False,
                "0844d4ffea6ffadf1b1609efeb3cd16a",
            ),
            (
                1,
                "Fender",
                "Telecaster",
                dt_utc(2021, 7, 1, 11),
                dt_utc(2021, 7, 1, 11),
                max_time,
                True,
                "d26e66604eb633097990f29a77043ba6",
            ),
            (
                2,
                "Gibson",
                "Les Paul",
                dt_utc(2021, 7, 1, 11),
                dt_utc(2021, 7, 1, 11),
                max_time,
                True,
                "0edf5ded053216e1497eadc7106a4f9b",
            ),
            (
                3,
                "Ibanez",
                "RG",
                dt_utc(2021, 8, 1, 11),
                dt_utc(2021, 8, 1, 11),
                dt_utc(2021, 9, 1, 11),
                False,
                "13ca996bde85cf51639c897a62cc01de",
            ),
            (
                3,
                "Ibanez",
                "RG",
                dt_utc(2021, 9, 1, 11),
                dt_utc(2021, 9, 1, 11),
                dt_utc(2021, 10, 1, 11),
                False,
                "8aff38592b66182bce84144844d90f7e",
            ),
            (
                3,
                "Ibanez",
                "RG",
                dt_utc(2021, 10, 1, 11),
                dt_utc(2021, 10, 1, 11),
                max_time,
                True,
                "968701f7bb46f602862a561ea1f9e998",
            ),
        ]

        # Read the resulting DataFrame from the SCD2 table and sort it for comparison
        df_result = dh.read().sort("Id", "TimeCol")

        # Assert that the DataFrame from the SCD2 table matches the expected data
        self.assertDataframeMatches(df_result, expected_data=expected_data)

        # Additional assertions to verify the integrity and correctness
        # of the SCD2 process. These checks ensure that the 'iscurrent' flag is properly
        # managed and that the deduplication logic works as expected
        self.assertEqual(dh.read().where("iscurrent").count(), 3)
        self.assertEqual(dh.read().where("iscurrent and Id=1").count(), 1)
        self.assertEqual(dh.read().where("iscurrent and Id=2").count(), 1)
        self.assertEqual(dh.read().where("iscurrent and Id=3").count(), 1)
        self.assertEqual(dh.read().select("hashvalue").distinct().count(), 6)

        """
        Then comes some new data:

        | Id | Col1   | Col2       | TimeCol               |
        |---------|--------|------------|-----------------------|
        | 1       | Fender | Asomething | 2022-01-01 00:00:00   |
        | 1       | Fender | Another   | 2022-02-01 00:00:00   |
        | 3       | Ibanez | Anew       | 2020-07-01 10:00:00   |
        """

        # This will update the id=1 and id = 3 in the sink table
        # Also, the id=3 is a late arrival from 2020 (a year earlier than the existing)
        data_new = [
            (1, "Fender", "Asomething", dt_utc(2022, 1, 1)),
            (1, "Fender", "Another", dt_utc(2022, 2, 1)),
            (3, "Ibanez", "Anew", dt_utc(2020, 7, 1, 10)),
        ]
        df_new = Spark.get().createDataFrame(data=data_new, schema=schema_start)

        _join_cols = ["Id"]
        _time_col = "TimeCol"
        SCD2UpsertLoader(sink_handle=dh, join_cols=_join_cols, time_col=_time_col).save(
            df_new
        )

        self.assertEqual(dh.read().where("iscurrent").count(), 3)
        self.assertEqual(dh.read().where("iscurrent and Id=1").count(), 1)
        self.assertEqual(dh.read().where("iscurrent and Id=2").count(), 1)
        self.assertEqual(dh.read().where("iscurrent and Id=3").count(), 1)
        self.assertEqual(dh.read().select("hashvalue").distinct().count(), 9)

        """
        New data arrives:

        | Id | Col1       | Col2       | TimeCol               |
        |---------|------------|------------|-----------------------|
        | 4       | FancyBrand | HelloThere | 2023-01-01 00:00:00   |

        """

        # This will update the sink table data with a new id=4
        data_new = [
            (4, "FancyBrand", "HelloThere", dt_utc(2023, 1, 1)),
        ]
        df_new = Spark.get().createDataFrame(data=data_new, schema=schema_start)

        SCD2UpsertLoader(sink_handle=dh, join_cols=_join_cols, time_col=_time_col).save(
            df_new
        )

        self.assertEqual(dh.read().where("iscurrent").count(), 4)
        self.assertEqual(dh.read().where("iscurrent and Id=1").count(), 1)
        self.assertEqual(dh.read().where("iscurrent and Id=2").count(), 1)
        self.assertEqual(dh.read().where("iscurrent and Id=3").count(), 1)
        self.assertEqual(dh.read().where("iscurrent and Id=4").count(), 1)
        self.assertEqual(dh.read().select("hashvalue").distinct().count(), 10)

        """
        Then this data arrives:
        | Id | Col1       | Col2    | TimeCol               |
        |---------|------------|---------|-----------------------|
        | 4       | FancyBrand | Changed | 2024-01-01 00:00:00   |
        """

        # This will update the data with a id=4 again
        data_new = [
            (4, "FancyBrand", "Changed", dt_utc(2024, 1, 1)),
        ]
        df_new = Spark.get().createDataFrame(data=data_new, schema=schema_start)

        _join_cols = ["Id"]
        _time_col = "TimeCol"
        SCD2UpsertLoader(sink_handle=dh, join_cols=_join_cols, time_col=_time_col).save(
            df_new
        )

        self.assertEqual(dh.read().where("iscurrent").count(), 4)
        self.assertEqual(dh.read().where("iscurrent and Id=1").count(), 1)
        self.assertEqual(dh.read().where("iscurrent and Id=2").count(), 1)
        self.assertEqual(dh.read().where("iscurrent and Id=3").count(), 1)
        self.assertEqual(dh.read().where("iscurrent and Id=4").count(), 1)
        self.assertEqual(dh.read().select("hashvalue").distinct().count(), 11)

    def test_02_multiple_join_cols(self):
        """
        Test the SCD2UpsertLoader functionality.
        This involves setting up a test database and table,
        loading initial data, and then validating the
        Slowly Changing Dimension Type 2 (SCD2) behavior.
        """
        self._create_table()

        dh = DeltaHandle.from_tc("MyTbl")

        """
        First test takes this input data:
            | Id |   Col1   |    Col2    |      TimeCol      |
            |---------|----------|------------|------------------|
            |    1    |  Fender  | Telecaster | 2021-07-01 10:00 |
            |    1    |  Fender  | Telecaster | 2021-07-01 10:00 |
            |    1    |  Super  | Telecaster | 2021-07-01 11:00 |
            |    2    |  Gibson  |  Les Paul  | 2021-07-01 11:00 |
            |    3    |  Ibanez  |     RG     | 2021-08-01 11:00 |
            |    3    |  Ibanez  |     RG     | 2021-09-01 11:00 |
            |    3    |  Ibanez  |     RG     | 2021-10-01 11:00 |


        """

        data_start = [
            (1, "Fender", "Telecaster", dt_utc(2021, 7, 1, 10)),
            # Duplicate when comparing the columns
            (1, "Fender", "Telecaster", dt_utc(2021, 7, 1, 10)),
            (1, "Super", "Telecaster", dt_utc(2021, 7, 1, 11)),
            (2, "Gibson", "Les Paul", dt_utc(2021, 7, 1, 11)),
            (3, "Ibanez", "RG", dt_utc(2021, 8, 1, 11)),
            (3, "Ibanez", "RG", dt_utc(2021, 9, 1, 11)),
            (3, "Ibanez", "RG", dt_utc(2021, 10, 1, 11)),
        ]
        schema_start = StructType(
            [
                StructField("Id", IntegerType(), True),
                StructField("Col1", StringType(), True),
                StructField("Col2", StringType(), True),
                StructField("TimeCol", TimestampType(), True),
            ]
        )
        # Create a DataFrame from the test data
        df_in = Spark.get().createDataFrame(data=data_start, schema=schema_start)

        # Define the columns for joining and time column for the SCD2 process
        _join_cols = ["Id", "Col1"]
        _time_col = "TimeCol"

        SCD2UpsertLoader(sink_handle=dh, join_cols=_join_cols, time_col=_time_col).save(
            df_in
        )

        max_time = datetime(2262, 4, 11, tzinfo=timezone.utc)

        """
        And generates this:
                |PK|Col1| Col2  |      Time     | ValFrom  |  ValTo   |Curr|Hash|
                |--|----|-------|---------------|----------|---------|----|----|
                |1 |Fend|Telec. |2021-07-01 10:|2021-07-01|max_time| T  |b69x|
                |1 |Super|Telec. |2021-07-01 11:|2021-07-01| max_time | T  |276x|
                |2 |Gibs|L.Paul|2021-07-01 11:|2021-07-01| max_time | T  |f79x|
                |3 |Iban| RG   |2021-08-01 11:|2021-08-01|21-09-01 1| F  |7d5x|
                |3 |Iban| RG   |2021-09-01 11:|2021-09-01|21-10-01 1| F  |274x|
                |3 |Iban| RG   |2021-10-01 11:|2021-10-01| max_time | T  |fb1x|
        """

        # Define the expected outcome after the initial data load
        # This includes the handling of duplicate records and
        # setting up the initial state of the SCD2 table

        expected_data = [
            (
                1,
                "Fender",
                "Telecaster",
                dt_utc(2021, 7, 1, 10),
                dt_utc(2021, 7, 1, 10),
                max_time,
                True,
                "0844d4ffea6ffadf1b1609efeb3cd16a",
            ),
            (
                1,
                "Super",
                "Telecaster",
                dt_utc(2021, 7, 1, 11),
                dt_utc(2021, 7, 1, 11),
                max_time,
                True,
                "8ee439f432890edef5e529b813a91f10",
            ),
            (
                2,
                "Gibson",
                "Les Paul",
                dt_utc(2021, 7, 1, 11),
                dt_utc(2021, 7, 1, 11),
                max_time,
                True,
                "0edf5ded053216e1497eadc7106a4f9b",
            ),
            (
                3,
                "Ibanez",
                "RG",
                dt_utc(2021, 8, 1, 11),
                dt_utc(2021, 8, 1, 11),
                dt_utc(2021, 9, 1, 11),
                False,
                "13ca996bde85cf51639c897a62cc01de",
            ),
            (
                3,
                "Ibanez",
                "RG",
                dt_utc(2021, 9, 1, 11),
                dt_utc(2021, 9, 1, 11),
                dt_utc(2021, 10, 1, 11),
                False,
                "8aff38592b66182bce84144844d90f7e",
            ),
            (
                3,
                "Ibanez",
                "RG",
                dt_utc(2021, 10, 1, 11),
                dt_utc(2021, 10, 1, 11),
                max_time,
                True,
                "968701f7bb46f602862a561ea1f9e998",
            ),
        ]

        # Read the resulting DataFrame from the SCD2 table and sort it for comparison
        df_result = dh.read().sort("Id", "TimeCol")

        # Assert that the DataFrame from the SCD2 table matches the expected data
        self.assertDataframeMatches(df_result, expected_data=expected_data)

        # Additional assertions to verify the integrity and correctness
        # of the SCD2 process. These checks ensure that the 'iscurrent' flag is properly
        # managed and that the deduplication logic works as expected
        self.assertEqual(dh.read().where("iscurrent").count(), 4)
        self.assertEqual(
            dh.read().where('iscurrent and Id=1 and Col1="Fender"').count(), 1
        )
        self.assertEqual(
            dh.read().where('iscurrent and Id=1 and Col1="Super"').count(), 1
        )
        self.assertEqual(dh.read().where("iscurrent and Id=2").count(), 1)
        self.assertEqual(dh.read().where("iscurrent and Id=3").count(), 1)
        self.assertEqual(dh.read().select("hashvalue").distinct().count(), 6)

    def test_03_streaming(self):
        """
        Test the SCD2UpsertLoader functionality
        in a streaming setup.
        """
        tc = Configurator()
        tc.set_debug()

        # Register database and tables
        self._register_database()
        self._register_stream_tables()

        # Initialize handlers to point to tables
        dh_source = DeltaHandle.from_tc("MyTblStreamSource")
        dh_sink = DeltaHandle.from_tc("MyTblStreamSink")

        # Recreate database
        dbh = DbHandle.from_tc("MyDb")
        dbh.drop_cascade()  # drop from last
        dbh.create()

        data_start = [
            (1, "Fender", "Telecaster", dt_utc(2021, 7, 1, 10)),
            # Duplicate when comparing the columns
            (1, "Fender", "Telecaster", dt_utc(2021, 7, 1, 10)),
            (1, "Fender", "Telecaster", dt_utc(2021, 7, 1, 11)),
            (2, "Gibson", "Les Paul", dt_utc(2021, 7, 1, 11)),
            (3, "Ibanez", "RG", dt_utc(2021, 8, 1, 11)),
            (3, "Ibanez", "RG", dt_utc(2021, 9, 1, 11)),
            (3, "Ibanez", "RG", dt_utc(2021, 10, 1, 11)),
        ]
        schema_start = StructType(
            [
                StructField("Id", IntegerType(), True),
                StructField("Col1", StringType(), True),
                StructField("Col2", StringType(), True),
                StructField("TimeCol", TimestampType(), True),
            ]
        )
        # Create a DataFrame from the test data
        df_in = Spark.get().createDataFrame(data=data_start, schema=schema_start)

        Spark.get().sql(
            f"""
                    CREATE TABLE {dh_source.get_tablename()}
                    (
                    Id integer,
                    Col1 string,
                    Col2 string,
                    TimeCol timestamp
                    )
                    LOCATION '{Configurator().get("MyTblStreamSource", "path")}'
                """
        )

        Spark.get().sql(
            f"""
                    CREATE TABLE {dh_sink.get_tablename()}
                    (
                    Id integer,
                    Col1 string,
                    Col2 string,
                    TimeCol timestamp,
                    ValidFrom timestamp,
                    ValidTo timestamp,
                    IsCurrent boolean,
                    HashValue string
                    )
                    LOCATION '{Configurator().get("MyTblStreamSink", "path")}'
                """
        )

        # Define the columns for joining and time column for the SCD2 process
        _join_cols = ["Id"]
        _time_col = "TimeCol"

        dh_source.overwrite(df_in)
        self.assertEqual(dh_source.read().count(), 7)

        o = Orchestrator()
        o.extract_from(StreamExtractor(dh_source, dataset_key="MyTblSource"))
        o.load_into(
            StreamLoader(
                loader=SCD2UpsertLoader(
                    dh_sink, join_cols=_join_cols, time_col=_time_col
                ),
                await_termination=True,
                checkpoint_path=Configurator().get(
                    "MyTblStreamSink", "checkpoint_path"
                ),
                query_name=Configurator().get("MyTblStreamSink", "query_name"),
            )
        )
        o.execute()

        max_time = datetime(2262, 4, 11, tzinfo=timezone.utc)

        """
        And generates this:
                |PK|Col1| Col2  |      Time     | ValFrom  |  ValTo   |Curr|Hash|
                |--|----|-------|---------------|----------|---------|----|----|
                |1 |Fend|Telec. |2021-07-01 10:|2021-07-01|21-07-01 1| F  |b69x|
                |1 |Fend|Telec. |2021-07-01 11:|2021-07-01| max_time | T  |276x|
                |2 |Gibs|L.Paul|2021-07-01 11:|2021-07-01| max_time | T  |f79x|
                |3 |Iban| RG   |2021-08-01 11:|2021-08-01|21-09-01 1| F  |7d5x|
                |3 |Iban| RG   |2021-09-01 11:|2021-09-01|21-10-01 1| F  |274x|
                |3 |Iban| RG   |2021-10-01 11:|2021-10-01| max_time | T  |fb1x|
        """

        # Define the expected outcome after the initial data load
        # This includes the handling of duplicate records and
        # setting up the initial state of the SCD2 table

        expected_data = [
            (
                1,
                "Fender",
                "Telecaster",
                dt_utc(2021, 7, 1, 10),
                dt_utc(2021, 7, 1, 10),
                dt_utc(2021, 7, 1, 11),
                False,
                "0844d4ffea6ffadf1b1609efeb3cd16a",
            ),
            (
                1,
                "Fender",
                "Telecaster",
                dt_utc(2021, 7, 1, 11),
                dt_utc(2021, 7, 1, 11),
                max_time,
                True,
                "d26e66604eb633097990f29a77043ba6",
            ),
            (
                2,
                "Gibson",
                "Les Paul",
                dt_utc(2021, 7, 1, 11),
                dt_utc(2021, 7, 1, 11),
                max_time,
                True,
                "0edf5ded053216e1497eadc7106a4f9b",
            ),
            (
                3,
                "Ibanez",
                "RG",
                dt_utc(2021, 8, 1, 11),
                dt_utc(2021, 8, 1, 11),
                dt_utc(2021, 9, 1, 11),
                False,
                "13ca996bde85cf51639c897a62cc01de",
            ),
            (
                3,
                "Ibanez",
                "RG",
                dt_utc(2021, 9, 1, 11),
                dt_utc(2021, 9, 1, 11),
                dt_utc(2021, 10, 1, 11),
                False,
                "8aff38592b66182bce84144844d90f7e",
            ),
            (
                3,
                "Ibanez",
                "RG",
                dt_utc(2021, 10, 1, 11),
                dt_utc(2021, 10, 1, 11),
                max_time,
                True,
                "968701f7bb46f602862a561ea1f9e998",
            ),
        ]

        # Read the resulting DataFrame from the SCD2 table and sort it for comparison
        df_result = dh_sink.read().sort("Id", "TimeCol")

        # Assert that the DataFrame from the SCD2 table matches the expected data
        self.assertDataframeMatches(df_result, expected_data=expected_data)

        # Additional assertions to verify the integrity and correctness
        # of the SCD2 process. These checks ensure that the 'iscurrent' flag is properly
        # managed and that the deduplication logic works as expected
        self.assertEqual(dh_sink.read().where("iscurrent").count(), 3)
        self.assertEqual(dh_sink.read().where("iscurrent and Id=1").count(), 1)
        self.assertEqual(dh_sink.read().where("iscurrent and Id=2").count(), 1)
        self.assertEqual(dh_sink.read().where("iscurrent and Id=3").count(), 1)
        self.assertEqual(dh_sink.read().select("hashvalue").distinct().count(), 6)

    def _create_table(self):
        # Set up test configurations for database and table
        tc = Configurator()
        tc.set_debug()
        self._register_database()

        tc.register(
            "MyTbl",
            {
                "name": "TestDb{ID}.TestTbl",
                "path": "/mnt/spetlr/silver/testdb{ID}/testtbl",
                "format": "delta",
            },
        )

        # Initialize and create the database and table for testing
        dbh = DbHandle.from_tc("MyDb")
        dbh.drop_cascade()  # drop from last
        dbh.create()
        dh = DeltaHandle.from_tc("MyTbl")

        Spark.get().sql(
            f"""
                    CREATE TABLE {dh.get_tablename()}
                    (
                    Id integer,
                    Col1 string,
                    Col2 string,
                    TimeCol timestamp,
                    ValidFrom timestamp,
                    ValidTo timestamp,
                    IsCurrent boolean,
                    HashValue string
                    )
                """
        )

        """
        First test takes this input data:
            | Id |   Col1   |    Col2    |      TimeCol      |
            |---------|----------|------------|------------------|
            |    1    |  Fender  | Telecaster | 2021-07-01 10:00 |
            |    1    |  Fender  | Telecaster | 2021-07-01 10:00 |
            |    1    |  Super  | Telecaster | 2021-07-01 11:00 |
            |    2    |  Gibson  |  Les Paul  | 2021-07-01 11:00 |
            |    3    |  Ibanez  |     RG     | 2021-08-01 11:00 |
            |    3    |  Ibanez  |     RG     | 2021-09-01 11:00 |
            |    3    |  Ibanez  |     RG     | 2021-10-01 11:00 |


        """

    def _register_database(self):
        tc = Configurator()
        tc.set_debug()
        tc.register(
            "MyDb", {"name": "TestDb{ID}", "path": "/mnt/spetlr/silver/testdb{ID}"}
        )

    def _register_stream_tables(self):
        tc = Configurator()
        tc.set_debug()
        tc.register(
            "MyTblStreamSource",
            {
                "name": "TestDb{ID}.MyTblStreamSource",
                "path": "/mnt/spetlr/silver/testdb{ID}/testtblstreamsource",
                "format": "delta",
                "checkpoint_path": "/mnt/spetlr/silver/testdb{ID}/"
                "_checkpoint_path_tblstreamsource",
                "query_name": "testquerytblstreamsource{ID}",
            },
        )

        tc.register(
            "MyTblStreamSink",
            {
                "name": "TestDb{ID}.TestTblStreamSink",
                "path": "/mnt/spetlr/silver/testdb{ID}/testtblstreamsink",
                "format": "delta",
                "checkpoint_path": "/mnt/spetlr/silver/testdb{ID}/"
                "_checkpoint_path_tblstreamsink",
                "query_name": "testquerytblstreamsink{ID}",
            },
        )



from .clean_column_names_transformer import CleanColumnNamesTransformer  # noqa: F401
from .country_to_alphacode_transformer import (  # noqa: F401
    CountryToAlphaCodeTransformer,
)
from .drop_oldest_duplicate_transformer import (  # noqa: F401
    DropOldestDuplicatesTransformer,
)
from .dropColumnsTransformer import DropColumnsTransformer  # noqa: F401
from .generate_md5_column_transformer import GenerateMd5ColumnTransformer  # noqa: F401
from .join_dataframes_transformer import JoinDataframesTransformer  # noqa: F401
from .select_and_cast_columns_transformer import (  # noqa: F401
    SelectAndCastColumnsTransformer,
)
from .selectColumnsTransformer import SelectColumnsTransformer  # noqa: F401
from .simple_dataframe_filter_transformer import (  # noqa: F401
    DataFrameFilterTransformer,
)
from .simple_sql_transformer import SimpleSqlServerTransformer  # noqa: F401
from .timezone_transformer import TimeZoneTransformer  # noqa: F401
from .union_transformer import UnionTransformer  # noqa: F401
from .validfromto_transformer import ValidFromToTransformer  # noqa: F401



from typing import List, Union

from pyspark.sql import DataFrame

from spetlr.etl import Loader
from spetlr.etl.loaders import Appendable, Overwritable, Upsertable


class SimpleLoader(Loader):
    def __init__(
        self,
        handle: Union[Overwritable, Appendable, Upsertable],
        *,
        mode: str = "overwrite",
        join_cols: List[str] = None,
        dataset_input_keys: List[str] = None,
        mergeSchema: bool = None,
        overwriteSchema: bool = None,
    ):
        super().__init__(dataset_input_keys=dataset_input_keys)
        self.mode = mode.lower()
        self.handle = handle
        self.join_cols = join_cols
        self.overwriteSchema = overwriteSchema
        self.mergeSchema = mergeSchema

    def save(self, df: DataFrame) -> None:
        if self.mode == "overwrite":
            self.handle.overwrite(
                df, overwriteSchema=self.overwriteSchema, mergeSchema=self.mergeSchema
            )
        elif self.mode == "upsert":
            self.handle.upsert(df, self.join_cols)
        else:
            self.handle.append(df, mergeSchema=self.mergeSchema)



import unittest

import pyspark.sql.types as T
from spetlrtools.testing import DataframeTestCase
from spetlrtools.time import dt_utc

from spetlr.spark import Spark
from spetlr.utils import SelectAndCastColumns


class TestSelectAndCastColumns(DataframeTestCase):
    @classmethod
    def setUpClass(cls):
        if Spark._spark is None:
            Spark.config("spark.sql.shuffle.partitions", 1)
            Spark.master("local[1]")

    def test_select_transformer(self):
        inputSchema = T.StructType(
            [
                T.StructField("value1", T.StringType(), True),
                T.StructField("value2", T.StringType(), True),
            ]
        )

        inputData = [
            (
                "value1",
                "value2",
            )
        ]

        input_df = Spark.get().createDataFrame(data=inputData, schema=inputSchema)

        transformer_schema = T.StructType(
            [
                T.StructField("value2", T.StringType(), True),
                T.StructField("value3", T.StringType(), True),
            ]
        )

        transformed_df = SelectAndCastColumns(df=input_df, schema=transformer_schema)

        expectedData = [
            (
                "value2",
                None,
            )
        ]

        self.assertEqualSchema(transformed_df.schema, transformer_schema)
        self.assertDataframeMatches(transformed_df, None, expectedData)

    def test_select_transformer_with_case_insensitive_matching(self):
        inputSchema = T.StructType(
            [
                T.StructField("VALUE1", T.StringType(), True),
                T.StructField("value2", T.StringType(), True),
            ]
        )

        inputData = [
            (
                "value1",
                "value2",
            )
        ]

        input_df = Spark.get().createDataFrame(data=inputData, schema=inputSchema)

        transformer_schema = T.StructType(
            [
                T.StructField("value1", T.StringType(), True),
                T.StructField("VALUE2", T.StringType(), True),
            ]
        )

        transformed_df = SelectAndCastColumns(
            df=input_df, schema=transformer_schema, caseInsensitiveMatching=True
        )

        expectedData = [
            (
                "value1",
                "value2",
            )
        ]

        self.assertEqualSchema(transformed_df.schema, transformer_schema)
        self.assertDataframeMatches(transformed_df, None, expectedData)

    def test_cast_transformer_valid_type_cast(self):
        datetime_to_use = dt_utc()

        inputSchema = T.StructType(
            [
                T.StructField("IntToStr", T.IntegerType(), True),
                T.StructField("StrToInt", T.StringType(), True),
                T.StructField("IntToDouble", T.IntegerType(), True),
                T.StructField("DoubleToInt", T.DoubleType(), True),
                T.StructField("IntToBooleanTrue", T.IntegerType(), True),
                T.StructField("IntToBooleanFalse", T.IntegerType(), True),
                T.StructField("BooleantoIntTrue", T.BooleanType(), True),
                T.StructField("BooleanToIntFalse", T.BooleanType(), True),
                T.StructField("StrToTimestamp", T.StringType(), True),
                T.StructField("TimestampToStr", T.TimestampType(), True),
            ]
        )

        inputData = [
            (
                42,
                "42",
                42,
                42.5,
                1,
                0,
                True,
                False,
                datetime_to_use.isoformat(),
                datetime_to_use,
            )
        ]

        input_df = Spark.get().createDataFrame(data=inputData, schema=inputSchema)

        transformer_schema = T.StructType(
            [
                T.StructField("IntToStr", T.StringType(), True),
                T.StructField("StrToInt", T.IntegerType(), True),
                T.StructField("IntToDouble", T.DoubleType(), True),
                T.StructField("DoubleToInt", T.IntegerType(), True),
                T.StructField("IntToBooleanTrue", T.BooleanType(), True),
                T.StructField("IntToBooleanFalse", T.BooleanType(), True),
                T.StructField("BooleantoIntTrue", T.IntegerType(), True),
                T.StructField("BooleanToIntFalse", T.IntegerType(), True),
                T.StructField("StrToTimestamp", T.TimestampType(), True),
                T.StructField("TimestampToStr", T.StringType(), True),
            ]
        )

        transformed_df = SelectAndCastColumns(df=input_df, schema=transformer_schema)

        expectedData = [
            (
                "42",
                42,
                42.0,
                42,
                True,
                False,
                1,
                0,
                datetime_to_use,
                datetime_to_use.strftime("%Y-%m-%d %H:%M:%S.%f").rstrip("0"),
            )
        ]

        self.assertEqualSchema(transformed_df.schema, transformer_schema)
        self.assertDataframeMatches(transformed_df, None, expectedData)

    def test_cast_transformer_invalid_type_cast(self):
        inputSchema = T.StructType(
            [
                T.StructField("StrToInt", T.StringType(), True),
            ]
        )

        inputData = [("test",)]

        input_df = Spark.get().createDataFrame(data=inputData, schema=inputSchema)

        transformer_schema = T.StructType(
            [
                T.StructField("StrToInt", T.IntegerType(), True),
            ]
        )

        transformed_df = SelectAndCastColumns(df=input_df, schema=transformer_schema)

        expectedData = [(None,)]

        self.assertEqualSchema(transformed_df.schema, transformer_schema)
        self.assertDataframeMatches(transformed_df, None, expectedData)


if __name__ == "__main__":
    unittest.main()



from typing import List

import pyspark.sql.functions as F
from pyspark.sql import DataFrame
from pyspark.sql.types import IntegerType, StringType, StructField, StructType

from spetlr.etl.log import LogTransformer
from spetlr.spark import Spark


class NullLogTransformer(LogTransformer):
    """
    A subclass of LogTransformer that logs the number of null values in a column. Note
    that this transformer relies on two separate .count() calls, meaning it can be an
    expensive operation if the input dataframe(s) have many rows.

    Arguments:
        log_name (str): The name for the log entry.
        column_name (str, optional): The specific column name for which the
            measurement is taken.
        dataset_input_keys (List[str], optional): A list of input dataset keys
            that this transformer will process. Defaults to None.
        dataset_output_key (str, optional): The key for the output dataset
            generated by this transformer. If not provided, a random UUID will
            be used as the default value. It is not required for when using the
            LogOrchestrator.
        consume_inputs (bool, optional): Flag to control dataset consuming behavior.
            Defaults to True
    """

    def __init__(
        self,
        *,
        log_name: str,
        column_name: str,
        dataset_input_keys: List[str] = None,
        consume_inputs: bool = True,
    ):
        super().__init__(
            log_name=log_name,
            dataset_input_keys=dataset_input_keys,
            consume_inputs=consume_inputs,
        )
        self.column_name = column_name

    def log(self, df: DataFrame) -> DataFrame:
        n_rows = df.count()
        n_nulls = df.filter(F.col(self.column_name).isNull()).count()
        percentage_of_nulls = round(n_nulls / n_rows, 4)

        percentage_of_nulls_formatted = "{:.4f}".format(percentage_of_nulls)

        schema = StructType(
            [
                StructField("Count", IntegerType(), True),
                StructField("NumberOfNulls", IntegerType(), True),
                StructField("PercentageOfNulls", StringType(), True),
                StructField("ColumnName", StringType(), True),
            ]
        )

        data = [
            (n_rows, n_nulls, percentage_of_nulls_formatted, self.column_name),
        ]

        return Spark.get().createDataFrame(schema=schema, data=data)



import pyspark.sql.types as T
from spetlrtools.testing import DataframeTestCase

from spetlr.spark import Spark
from spetlr.transformers import SelectColumnsTransformer


class TestSelectColumnsTransformer(DataframeTestCase):
    def test_select_columns_transformer(self):
        inputSchema = T.StructType(
            [
                T.StructField("Col1", T.StringType(), True),
                T.StructField("Col2", T.IntegerType(), True),
                T.StructField("Col3", T.DoubleType(), True),
                T.StructField("Col4", T.StringType(), True),
                T.StructField("Col5", T.StringType(), True),
            ]
        )
        inputData = [("Col1Data", 42, 13.37, "Col4Data", "Col5Data")]

        input_df = Spark.get().createDataFrame(data=inputData, schema=inputSchema)

        expectedData = [(42, 13.37, "Col4Data")]

        transformed_df = SelectColumnsTransformer(
            columnList=["Col2", "Col3", "Col4"]
        ).process(input_df)

        self.assertDataframeMatches(transformed_df, None, expectedData)



import unittest

from spetlr import Configurator
from spetlr.configurator.sql.parse_sql import parse_sql_code_to_config
from spetlr.deltaspec.DeltaDatabaseSpec import DeltaDatabaseSpec
from spetlr.spark import Spark


class TestDbSpec(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        c = Configurator()
        c.clear_all_configurations()
        c.set_debug()

    def test_DbSpecFromDisk(self):
        c = Configurator()
        master_sql = """
            -- SPETLR.CONFIGURATOR key: MyDbAlias
            CREATE DATABASE IF NOT EXISTS my_db_name{ID}
            COMMENT "Really great db"
            LOCATION "/tmp/my_db{ID}";
            """
        config = parse_sql_code_to_config(master_sql)

        k, v = config.popitem()
        self.assertEqual(k, "MyDbAlias")
        c.register(k, v)
        db_parsed = DeltaDatabaseSpec.from_tc(k)
        Spark.get().sql(master_sql.format(**c.get_all_details()))
        db_read = DeltaDatabaseSpec.from_spark(db_parsed.fully_substituted().name)

        self.assertEqual(db_parsed.fully_substituted(), db_read)

        Spark.get().sql(f"DROP DATABASE {db_parsed.fully_substituted().name} CASCADE")



from typing import List

from pyspark.sql import DataFrame

from spetlr.sql.CommonBaseServer import CommonBaseServer


class SqlBaseServer(CommonBaseServer):
    def load_sql(self, sql: str) -> DataFrame:
        pass

    def upsert_to_table_by_name(
        self,
        df_source: DataFrame,
        table_name: str,
        join_cols: List[str],
    ):
        pass

    def truncate_table_by_name(self, table_name: str) -> None:
        pass

    def drop_table_by_name(self, table_name: str) -> None:
        pass



from typing import Protocol

from pyspark.sql import DataFrame

from spetlr.etl.extractors.lazy_extractor import LazyExtractor


class Readable(Protocol):
    def read(self) -> DataFrame:
        pass


class LazySimpleExtractor(LazyExtractor):
    """This extractor will extract from any object that has a .read() method."""

    def __init__(self, handle: Readable, dataset_key: str = None):
        super().__init__(dataset_key=dataset_key)
        self.handle = handle

    def read(self) -> DataFrame:
        return self.handle.read()
